// ======================================================================
//
// Vector.cpp
// jeff grills
//
/// Vector/Functions
///
/// This class is designed to be very efficient.  Routines that typically generate temporary variables
/// have not been implemented, in particular the overloaded operators such as add and subtract.
///
// copyright 1998 origin systems inc
//
// ======================================================================

#include "First.h"
#include "Vector.h"

#include "Random.h"

// =========================================================================

const real   Vector::normalizeThreshold = CONST_REAL(0.0001);

const Vector Vector::zero(CONST_REAL(0), CONST_REAL(0), CONST_REAL(0));
const Vector Vector::unitX(CONST_REAL(1), CONST_REAL(0), CONST_REAL(0));
const Vector Vector::unitY(CONST_REAL(0), CONST_REAL(1), CONST_REAL(0));
const Vector Vector::unitZ(CONST_REAL(0), CONST_REAL(0), CONST_REAL(1));
const Vector Vector::unitNegativeX(CONST_REAL(-1), CONST_REAL(0),  CONST_REAL(0));
const Vector Vector::unitNegativeY(CONST_REAL(0),  CONST_REAL(-1), CONST_REAL(0));
const Vector Vector::unitNegativeZ(CONST_REAL(0),  CONST_REAL(0),  CONST_REAL(-1));

// =========================================================================
/// Vector/Functions/Misc/bool normalize(void)
///
/// Normalizes the vector's length to one
///
/// Return Value:
///
/// True if sucessful, false if not
///
/// Remarks:
///
/// Always check the return value of this routine.  Vectors with very small magnitudes
/// cannot be normalized without causing a division by zero.

bool Vector::normalize(void)
{
	real mag;

	mag = magnitude();
	if (mag >= normalizeThreshold)
	{
		mag = RECIP(mag);
		x *= mag;
		y *= mag;
		z *= mag;
		return true;
	}

	return false;
}

// ----------------------------------------------------------------------
/// Vector/Functions/Math/void dot(const Vector *v, real *dotProducts, int cout) const
///
/// Computes the dot product between this vector and an array of other vectors
///
/// Parameters:
///
/// v Pointer to the array of vectors to compute the dot product against
///
/// dotProducts Array to store the resulting dot products
///
/// count Number of elements in the each array
///
/// Remarks:
///
/// This version of this routine computes many dot products at one.  It is very useful
/// for computing lighting and similiar valuess.

void Vector::dot(const Vector *v, real *dotProduct, int count) const
{
	for ( ; count--; v++, dotProduct++)
		*dotProduct =((x * v->x) + (y * v->y) + (z * v->z));
}

// ----------------------------------------------------------------------
/// Vector/Functions/Orientation/real theta(void) const
///
/// Computes angle of rotation around the Y axis.
///
/// Return Value:
///
/// The angle of rotation around the Y axis, in radians, specified from -PI to PI.
///
/// Remarks:
///
/// This routine is slow, because it calls atan2().

real Vector::theta(void) const
{
	return atan2(x, z);
}

// ----------------------------------------------------------------------
/// Vector/Functions/Orientation/real phi(void) const
///
/// Computes angle of rotation around the X axis
///
/// Return Value:
///
/// The angle of rotation around the X axis, in radians, specified from -PI/2 to PI/2.
///
/// Remarks:
///
/// The value returned is the pitch required after any necessary yaw has been applied.
/// Thus, the result may only be PI/2 radians in each direction.  
/// If this is not desired, call phiMajor() instead.
///
/// This routine is very slow, because it calls sqrt() and atan2().

real Vector::phi(void) const
{
	return atan2(-y, sqrt(sqr(x) + sqr(z)));
}

// ----------------------------------------------------------------------
/// Vector/Functions/Orientation/real phiMajor(void) const
///
/// Computes angle of rotation around the X axis
///
/// Return Value:
///
/// The angle of rotation around the X axis, in radians, specified from -PI to PI.
///
/// Remarks:
///
/// The value returned is the pitch required assuming no yaw will be applied.
/// Because no yaw is applied, the range of this function is PI in each direction.
/// If compensation for yaw is desired, call phi() instead.
///
/// This routine is slow, because it must call atan2().

real Vector::phiMajor(void) const
{
	return atan2(-y, z);
}

// ----------------------------------------------------------------------
/// Vector/Functions/Building/void makePerpendicular(const Vector &v)
///
/// Creates an arbitrary vector orthogonal to the specified vector.
///
/// Parameters:
///
/// v The vector to construct a perpendicular vector to
///
/// Remarks:
///
/// This routine will create a arbitrary orthogonal vector by using the cross product with
/// either the unit X vector or the unit Y vector, whichever will create a numerically better
/// result.
///
/// If the resultant vector's magnitude will be the same as the source vector's magnitude.

void Vector::makePerpendicular(const Vector &to)
{
	if (abs(to.dot(unitX)) < abs(to.dot(unitY)))
		cross(to, unitX);
	else
		cross(to, unitY);
}

// ----------------------------------------------------------------------
/// Vector/Functions/Magnitude/real approximateMagnitude(void) const
///
/// Return Value:
///
/// An approximation of the magnitude of this vector
/// 
/// Remarks:
///
/// This routine uses the formula (long + (mid + short) / 4) to compute the 
/// approximate magnitude of the vector.  Its error is well known, but I do
/// not have that information on hand.
///
/// See also magnitude()

real Vector::approximateMagnitude(void) const
{
	real a, b, c;

	a = abs(x);
	b = abs(y);
	c = abs(z);

	// get the max in c
	minmax(a, c);
	minmax(b, c);

	return (c + ((a + b) * CONST_REAL(0.25f)));
}

// ----------------------------------------------------------------------
/// Vector/Functions/Magnitude/real approximateMagnitudeBetween(const Vector &v) const
///
/// Returns an approximation of the magnitude of the vector between the point at this vector 
/// and the point at vector v
///
/// Return value:
///
/// An approximation of the magnitude of the vector between the point at this vector and the point at vector v
///
/// Parameters:
///
/// v Point to determine the approximate magnitude to
///
/// Remarks:
///
/// See also magnitudeBetween().

real Vector::approximateMagnitudeBetween(const Vector &v) const
{
	real a, b, c;

	a = abs(x - v.x);
	b = abs(y - v.y);
	c = abs(z - v.z);

	// get the max in c
	minmax(a, c);
	minmax(b, c);

	return (c + ((a + b) * CONST_REAL(0.25f)));
}

// -----------------------------------------------------------------------
/// Vector/Functions/Building/Vector &randomUnit(void)
///
/// Creates an unit vector with a random orientation
///
/// Return Value:
///
/// This vector, so it can be used in subsequent math operations
///
/// Remarks:
///
/// This routine returns a uniform distribution of unit vectors on the unit vector sphere.
///
/// This routine was copied from the comp.graphics.algorithms FAQ

Vector &Vector::randomUnit(void)
{
	real r, t;

	// pick a uniformally distributed point along the Z axis
	z = ::random(CONST_REAL(2)) - CONST_REAL(1);

	// pick a uniformally distributed point around a circle
	t = ::random(Pi_2);

	// compute the X and Y for the point on the circle at the z,theta
	r = sqrt(CONST_REAL(1) - sqr(z));
	x = r * cos(t);
	y = r * sin(t);

	return *this;
}

// ------------------------------------------------------------------------
/// Vector/Functions/Building/Vector &random(real radius)
///
/// Creates an unit vector with a random orientation and magnitude with the specified radius
///
/// Return Value:
///
/// This vector, so it can be used in subsequent math operations
///
/// Remarks:
///
/// This routine creates a uniformly distributed random vector within the sphere specified
/// by the radius.
///
/// Unfortunately, it does so by creating a uniformly distributed vector within a cube
/// specified by the radius, and then verifies that the vector is within the sphere.  If
/// it is not, it iterates this process until it is within the sphere.
///
/// We could certainly use a better implementation than this.

Vector &Vector::random(real radius)
{
	real s, t;

	// randomly pick vectors until we get one within the sphere
	t = radius * CONST_REAL(2);
	s = sqr(radius);
	do
	{
		x = ::random(t) - radius;
		y = ::random(t) - radius;
		z = ::random(t) - radius;
	} while (magnitudeSquared() > s);

	return *this;
}

// ======================================================================
